/*
 * vfread.c version 1.0 part of the vfloppy 1.2 package
 *
 * Copyright 2002 by Fred Jan Kraan (fjkraan@xs4all.nl)
 *
 * vfread is placed under the GNU General Public License in July 2002.
 *
 *  This file is part of Vfloppy 1.2.
 *
 *  Vfloppy is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Vfloppy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Vfloppy; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * vfread - reads directory and files from disk images
 *          used by the vfloppy package.
 * usage: vfread <imageName> [<fileName>]
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "vfread.h"
#include "../epsp.h"
#include "../fdc.h"

#define PERMS 0664

char *fileName;
char dirFileName[12];
char *imageName;
int fnd;
int imd;
unsigned char blockList[MAXBLOCKCOUNT];
unsigned char data[BLOCKSIZE];
int entryNo = -1;
int lastRecCnt;

/*
 - read arguments
 - check and mount image
 - filename to uppercase and convert to directory format
 - parse image directory and return block list
   abort on file not found.
 - create file and dump the blocks
 - close, unmount, exit
 */

void showDirEntry(unsigned char *dent)
{
	/*
	 * convert data from the directory entry *dent
	 * and represent it as a sortof directory list
	 */
        char filename[13];
	int localEntryNo, recCnt, blockCnt, localLastRecs;
        if (dent[DIRUSERNO] != FMTPATT)
	{
                memcpy(filename,&dent[DIRNAMEBASE],8);
                filename[8]='.';
                memcpy(&filename[9],&dent[DIREXTBASE],3);
                filename[12]=0;
		localEntryNo = ((32 * dent[DIRENTRYNOFIELD]) + dent[DIREXTCNTHIGH]) / (EXM + 1);
		recCnt = (dent[DIRENTRYNOFIELD] & EXM) * 128 + dent[DIRRECS];
		blockCnt = recCnt / RECSPERBLOCK;
		localLastRecs = recCnt % RECSPERBLOCK;
		if (localLastRecs == 0)
		{
			localLastRecs = 16;
		} else {
			blockCnt++;
		}
                printf("%2d:%s ", dent[DIRUSERNO], filename);
		printf("%3d records ", recCnt);
		printf("(%3d blks, %2d recs in last blk) ", blockCnt, localLastRecs);
                if (dent[DIRENTRYNOFIELD] != 0)
                        printf("dir. entry %d", localEntryNo);
                if (dent[DIRS2] != 0)
                        printf("  ** Corrupt Directory Entry **");
                printf("\n");
        } 
}

static int blockRead(int fd, unsigned char *data, unsigned long blockStart, int len)
{
        int err;
        if(lseek(fd, blockStart, 0)==-1)
        {
                perror("vfread: lseek");
                return -1;
        }
        err=read(fd,data,len);
        return err;
}

void showDir(int id)
{
	unsigned long dirEntry;
	int i;
	int err;

	/* - loop 64 entries
	 * - start at 0x00080, increment with 0x00020
	 */

	for (i = 0; i < DIRENTRYCOUNT; i++)
	{
		dirEntry = (DIRBASE + i * DIRINCR);
		err = blockRead(id, data, dirEntry, DIRINCR);
        	showDirEntry(data); 
	}
}

int mountImage(char *imName)
{
	if((imd = open(imName,O_RDONLY,0)) == -1)
	{
		fprintf(stderr," failed to open %s\n",imName);
		exit(1);
	}
	printf("Disk image %s (fd: %d) mounted ok.\n",imName,imd);
	return imd;
}

void convertFileName(unsigned char *dFName, unsigned char *flName) 
{
	/* conversie fileName dirFileName
	   - to uppercase
	   - add spaces to basename to 8 chars
	   - remove dot
	   - add spaces to extension to 3 chars
	 */
	int i = 0;
	int j = 0;
	int c;
	for (c = 0; c < 11; c++)
		dFName[c] = ' ';
	dFName[11] = '\0';

	/* walk though the array to fill and check the
	 * input array:
	 * - end of string: 	exit
	 * - a dot:		write the rest at the extension location
	 * Most problems are not checked here, but will hopefully fail
	 * nicely at filename comparison.
	 */
	for (i = 0; i < 12; i++)
	{
		if (flName[i] == '\0')
			break;
		if (flName[i] == '.')
		{
			j = 8;
		} else {
			if (j < 11)
			{
				dFName[j] = toupper(flName[i]);
				j++;
			}	
		}
	} 
/*	printf("output: \'%s\'\n", dFName); */
}

int cmpEntry(unsigned char *dent, char *dFlName)
{
	int localEntryNo = dent[DIRENTRYNOFIELD];
	int recCnt, localLastRecs;
/*	int i; */
	char cmpFileName[12];

	/*
	 * - get entry
	 * - if match: get 'dir entry no'
	 * - store blocklist in array at 'dir entry no' * 16
	 */

	memcpy(&cmpFileName[0], &dent[DIRNAMEBASE], DIRNAMESIZE);
	cmpFileName[DIRNAMESIZE] = '\0';
	if (strcmp(cmpFileName, dFlName) == 0)
	{
		printf("Found file \'%s\' entryNo %d\n", dFlName, dent[DIRENTRYNOFIELD]);
		/* dent[DIRENTRYNOFIELD] 0 and 1 both mean the first dir entry */
		if (localEntryNo > 0)
			localEntryNo--;
		/* selects the part of the BlockList the blocks are copied to */
/*		printf("Copied blocks to %d\n", (localEntryNo * DIRBLOCKSIZE)); */
		memcpy(&blockList[localEntryNo * DIRBLOCKSIZE], &dent[DIRBLOCKBASE], DIRBLOCKSIZE);
/*		printf("blk: ");
		for (i = 0; i < DIRBLOCKSIZE; i++)
		{
			printf("%d: %X = %X; ", i + localEntryNo * DIRBLOCKSIZE, dent[i + DIRBLOCKBASE], blockList[i + DIRBLOCKSIZE * localEntryNo]);
			if (blockList[i] == '\0')
				break;
		} 
		printf("\n"); */
		/* 
		 * determine the number of used records in the
		 * last allocated block. Done for all entries,
		 * but the globals contain the right data in the end.
		 */
		recCnt = (dent[DIRENTRYNOFIELD] & EXM) * 128 + dent[DIRRECS];
                localLastRecs = recCnt % RECSPERBLOCK;
                if (localLastRecs == 0)
                {
                        localLastRecs = DIRBLOCKSIZE;
                }

		if (localEntryNo > entryNo)
		{
			entryNo = localEntryNo;
			lastRecCnt = localLastRecs;
		}
        	return 1;
	} else {
		return 0;
	}
}

int findFile(int id, unsigned char *dFName)
{
	unsigned long dirEntry;
	int i, entryCnt = 0;
	int err;

	/* - loop 64 entries
	 * - start at 0x00080, increment with 0x00020
	 */

	for (i = 0; i < DIRENTRYCOUNT; i++)
	{
		dirEntry = (DIRBASE + i * DIRINCR);
		err = blockRead(id, data, dirEntry, DIRENTRYCOUNT);
/*		printf("comparing %s with block %ld\n",dFName,dirEntry); */
        	entryCnt = entryCnt + cmpEntry(data, dFName);
 	}
	printf("Found entries: %d\n", entryCnt);
	return entryCnt;
}

int createFile(unsigned char *fileName)
{
	int i, fd, blockSize;
	long unsigned int blkAddr;

	/*
	 * for each block until block number is 0:
	 * - translate block number to address in image
	 * - get block and write it to the file
	 */

	fd = open(fileName, O_WRONLY|O_CREAT, PERMS);
	if (fd < 0) 
	{
		printf("Error opening file %s\n", fileName);
		return 1;
	} 
	printf("Blocks: ");
	for (i = 0; i < 256; i++)
	{
		if (blockList[i] == '\0')
			break;
		blkAddr = BLOCKBASE + blockList[i] * BLOCKSIZE;
		printf("%X: %4X ", blockList[i], blkAddr);
		if (blockList[i + 1] == '\0')
		{
			blockSize = lastRecCnt * RECSIZE;
			printf("Last block records: %d", lastRecCnt); 
		} else {
			blockSize = BLOCKSIZE;
		}
		blockRead(imd, data, blkAddr, blockSize); 
		write(fd, data, blockSize); 
	}
	printf("\n");
	close(fd); 
	return 0;
}

int main(int argc, char *argv[])
{
	memset(&data[0], 0x0, BLOCKSIZE);
	memset(&blockList[0], 0x0, MAXBLOCKCOUNT);
	if(argc<2 || argc> 3)
	{
		fprintf(stderr, "%s <imageName> [<fileName>]\n", argv[0]);
		exit(1);
	}
	imageName = argv[1];
	imd = mountImage(imageName);

        if(argc == 3)
		fileName = argv[2];
	else
	{
		printf("To directory mode\n");
		showDir(imd);
		return 0;
	}
	convertFileName(dirFileName, fileName);
	printf("Converted name: \'%s\'\n", dirFileName);
	if (findFile(imd, dirFileName) > 0)
		createFile(fileName); 

	return 0;
}
